import collections
import itertools
import re
from typing import List, DefaultDict, Iterable

import psutil
from loguru import logger

ProcessFilterPredicate = collections.namedtuple(
    "ProcessFilterPredicate", ["name", "conditions"]
)


def _match_process(process: psutil.Process, predicate: str) -> bool:
    name, regex = predicate
    value = getattr(process, name)()
    if isinstance(value, int):
        value = str(value)
    elif isinstance(value, (list, tuple)):
        value = " ".join(value)

    return bool(re.match(regex, value))


def filter_net_connections(
    predicates: Iterable[ProcessFilterPredicate],
) -> DefaultDict[str, List[psutil._common.pconn]]:
    filtered: DefaultDict[str, List[psutil._common.pconn]] = collections.defaultdict(
        list
    )
    connections = psutil.net_connections()
    for connection, predicate in itertools.product(connections, predicates):
        # Stop no specified conditions from matching every process
        if not (predicate.conditions and connection.pid):
            continue

        try:
            process = psutil.Process(connection.pid)
            if all(
                _match_process(process, condition) for condition in predicate.conditions
            ):
                filtered[predicate.name].append(connection)
        except psutil.NoSuchProcess:
            logger.warning(
                "Process with PID {} died while filtering network connections",
                connection.pid,
            )
            continue

    return filtered



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['traffictoll']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.4.1,<0.5.0', 'psutil>=5.6.7,<6.0.0', 'ruamel.yamL>=0.16.6,<0.17.0']

entry_points = \
{'console_scripts': ['tt = traffictoll.__main__:main']}

setup_kwargs = {
    'name': 'traffictoll',
    'version': '1.0.2',
    'description': 'NetLimiter-like traffic shaping for Linux',
    'long_description': "# TrafficToll\nNetLimiter-like traffic shaping for Linux\n\n# Description\nTrafficToll allows you to limit download and upload bandwidth globally\n(per interface) and per process, even during the process' runtime.\n\nThe configuration can be easily adjusted and new limits applied at any\npoint, as opposed to similar tools which either can only apply fixed\nglobal limits to the interface, certain ports, or require you to start\nthe process through them (and thus restart the target process to change\nthe limits).\n\n# Usage\n`# tt device config`\n\nWhere `device` is the interface you want to limit (usually the one you\nconnect to the internet with). For example:\n\n* `# tt enp3s0 night.yaml --delay 0.5` (regular interface, check every\nhalf second for change in networked processes)\n* `# tt tun0 day.yaml --logging-level DEBUG` (VPN interface, adjust\nlogging level to DEBUG)\n\nCurrently TrafficToll works based on a YAML configuration file. The configuration file\nis best explained by example:\n\n```YAML\n# Global limits\ndownload: 500kbps\nupload: 100kbps\n\n# Matched process limits\nprocesses:\n  Vivaldi:\n    download: 100kbps\n    match:\n      - exe: /opt/vivaldi/vivaldi-bin\n\n  Discord:\n    download: 300kbps\n\n    # This won't work, the specified upload exceeds the global upload, it will\n    # be 100kb/s max\n    upload: 200kbps\n    match:\n      - exe: /opt/discord/Discord\n\n  JDownloader 2:\n    # JDownloader 2 obviously has its own traffic shaping, this is just here as\n    # an example to show that matching on something else than the executable's\n    # path is possible\n    download: 300kbps\n    match:\n      - cmdline: .* JDownloader.jar\n```\n\nUnits can be specified in all formats that `tc` supports, namely: bit \n(with and without suffix), kbit, mbit, gbit, tbit, bps, kbps, mbps,\ngbps, tbps. To specify in IEC units, replace the SI prefix (k-, m-, g-,\nt-) with IEC prefix (ki-, mi-, gi- and ti-) respectively.\n\nAll limits can be omitted, in which case obviously no limiting will be\napplied. A process is selected when all predicates in the match section\nmatch. Every attribute [`psutil.Process`](https://psutil.readthedocs.io/en/latest/index.html#psutil.Process)\nprovides on Linux can be matched on, using regular expressions.\n\nWhen you terminate `tt` using Ctrl+C all changes to the traffic\nscheduling will be reverted, allowing you to easily update the config\nand apply new limits.\n\n# Installation\n`$ pip install traffictoll`\n\n`tt` has to be run as root.\n\n# Screenshots\nBecause a picture is always nice, even for CLI applications:\n\n![](https://i.imgur.com/EsOla66.png)\n",
    'author': 'cryzed',
    'author_email': 'cryzed@googlemail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cryzed/TrafficToll',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import argparse
import atexit
import collections
import enum
import time
from typing import Dict, DefaultDict, Set

from loguru import logger
from ruamel.yaml import YAML

from .net import ProcessFilterPredicate, filter_net_connections
from .tc import (
    INGRESS_QDISC_PARENT_ID,
    tc_add_htb_class,
    tc_add_u32_filter,
    tc_remove_qdisc,
    tc_remove_u32_filter,
    tc_setup,
)

CONFIG_ENCODING = "UTF-8"


class _TrafficType(enum.Enum):
    Ingress = enum.auto()
    Egress = enum.auto()


def get_argument_parser() -> argparse.ArgumentParser:
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("device")
    argument_parser.add_argument("config")
    argument_parser.add_argument("--delay", "-d", type=float, default=1)
    argument_parser.add_argument(
        "--logging-level",
        "-l",
        choices={"TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"},
        default="INFO",
    )
    return argument_parser


def _clean_up(ingress_interface: str, egress_interface: str) -> None:
    logger.info("Cleaning up QDiscs")
    tc_remove_qdisc(ingress_interface)
    tc_remove_qdisc(egress_interface)
    tc_remove_qdisc(egress_interface, INGRESS_QDISC_PARENT_ID)


def main(arguments: argparse.Namespace) -> None:
    with open(arguments.config, "r", encoding=CONFIG_ENCODING) as file:
        config = YAML().load(file)

    # TODO: Parse download rate
    global_download_rate = config.get("download")
    global_upload_rate = config.get("upload")
    if global_download_rate:
        logger.info(
            "Setting up global download limiter with max rate {}", global_download_rate
        )
    if global_upload_rate:
        logger.info(
            "Setting up global upload limiter with max rate {}", global_upload_rate
        )

    ingress, egress = tc_setup(
        arguments.device, global_download_rate, global_upload_rate
    )
    ingress_interface, ingress_qdisc_id, ingress_root_class_id = ingress
    egress_interface, egress_qdisc_id, egress_root_class_id = egress

    atexit.register(_clean_up, ingress_interface, egress_interface)

    process_filter_predicates = []
    class_ids: Dict[_TrafficType, Dict[str, int]] = {
        _TrafficType.Ingress: {},
        _TrafficType.Egress: {},
    }
    for name, process in (config.get("processes", {}) or {}).items():
        # Prepare process filter predicates to match network connections
        conditions = [list(match.items())[0] for match in process.get("match", [])]
        if not conditions:
            logger.warning(
                "No conditions for {!r} specified, it will never be matched", name
            )

        predicate = ProcessFilterPredicate(name, conditions)
        process_filter_predicates.append(predicate)

        # Set up classes for download/upload limiting
        download_rate = process.get("download")
        upload_rate = process.get("upload")
        if download_rate:
            logger.info(
                "Setting up download limiter for {!r} with max rate {}",
                name,
                download_rate,
            )
            egress_class_id = tc_add_htb_class(
                ingress_interface,
                ingress_qdisc_id,
                ingress_root_class_id,
                download_rate,
            )
            class_ids[_TrafficType.Ingress][name] = egress_class_id
        if upload_rate:
            logger.info(
                "Setting up upload limiter for {!r} with max rate {}", name, upload_rate
            )
            ingress_class_id = tc_add_htb_class(
                egress_interface, egress_qdisc_id, egress_root_class_id, upload_rate
            )
            class_ids[_TrafficType.Egress][name] = ingress_class_id

    port_to_filter_id: Dict[_TrafficType, Dict[int, str]] = {
        _TrafficType.Ingress: {},
        _TrafficType.Egress: {},
    }

    def add_ingress_filter(port: int, class_id: int) -> None:
        filter_id = tc_add_u32_filter(
            ingress_interface,
            f"match ip dport {port} 0xffff",
            ingress_qdisc_id,
            class_id,
        )
        port_to_filter_id[_TrafficType.Ingress][port] = filter_id

    def add_egress_filter(port: int, class_id: int) -> None:
        filter_id = tc_add_u32_filter(
            egress_interface, f"match ip sport {port} 0xffff", egress_qdisc_id, class_id
        )
        port_to_filter_id[_TrafficType.Egress][port] = filter_id

    def remove_filters(port: int) -> None:
        ingress_filter_id = port_to_filter_id[_TrafficType.Ingress].get(port)
        if ingress_filter_id:
            tc_remove_u32_filter(ingress_interface, ingress_filter_id, ingress_qdisc_id)
            del port_to_filter_id[_TrafficType.Ingress][port]

        egress_filter_id = port_to_filter_id[_TrafficType.Egress].get(port)
        if egress_filter_id:
            tc_remove_u32_filter(egress_interface, egress_filter_id, egress_qdisc_id)
            del port_to_filter_id[_TrafficType.Egress][port]

    filtered_ports: DefaultDict[str, Set[int]] = collections.defaultdict(set)
    while True:
        filtered_connections = filter_net_connections(process_filter_predicates)
        for name, connections in filtered_connections.items():
            ports = set(connection.laddr.port for connection in connections)
            ingress_class_id = class_ids[_TrafficType.Ingress].get(name)
            egress_class_id = class_ids[_TrafficType.Egress].get(name)

            # Add new port filters
            new_ports = sorted(ports.difference(filtered_ports[name]))
            if new_ports:
                logger.info(
                    "Filtering traffic for {!r} on local ports {}",
                    name,
                    ", ".join(map(str, new_ports)),
                )
                for port in new_ports:
                    if ingress_class_id:
                        add_ingress_filter(port, ingress_class_id)
                    if egress_class_id:
                        add_egress_filter(port, egress_class_id)

            # Remove old port filters
            freed_ports = sorted(filtered_ports[name].difference(ports))
            if freed_ports:
                logger.info(
                    "Removing filters for {!r} on local ports {}",
                    name,
                    ", ".join(map(str, freed_ports)),
                )
                for port in freed_ports:
                    remove_filters(port)

            filtered_ports[name] = ports

        # Remove freed ports for unmatched processes (process died or predicate
        # conditions stopped matching)
        for name in set(filtered_ports).difference(filtered_connections):
            freed_ports = sorted(filtered_ports[name])
            if freed_ports:
                logger.info(
                    "Removing filters for {!r} on local ports {}",
                    name,
                    ", ".join(map(str, freed_ports)),
                )
                for port in freed_ports:
                    remove_filters(port)
            del filtered_ports[name]

        time.sleep(arguments.delay)
